"""------------------------------------------------------------------------------------------
Bayesian Affect Control Theory
Interactive Example
Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey
September 2013
Use for research purposes only.
Please do not re-distribute without written permission from the author
Any commerical uses strictly forbidden.
Code is provided without any guarantees.
Research sponsored by the Natural Sciences and Engineering Council of Canada (NSERC).
use python2.6
see README for details
----------------------------------------------------------------------------------------------"""

from bayesact import *

NP.set_printoptions(precision=5)
NP.set_printoptions(suppress=True)
NP.set_printoptions(linewidth=10000)
 


#-----------------------------------------------------------------------------------------------------------------------------
#user-defined parameters 
#-----------------------------------------------------------------------------------------------------------------------------
#agent knowledge of client id: 
#0 : nothing
#1 : one of a selection of  num_confusers+1 randoms
#2 : exactly - use this to mimic interact
#3 : same as 0 but also agent does not know its own id
agent_knowledge=2

# agent gender
agent_gender="male"

# client gender
client_gender="male"

#possibly set the agent id to be something
agent_id="tutor"
#if not in database (including "") then it is a randomly drawn id
#agent_id=""

#can also set the client id here if agent_knowledge = 2 (knows id of client - see above)
#if agent_knowledge is 0 then this is ignored
client_id = "student"
#client_id = ""

#how often do we want to see the full id sets learned by the agent
get_full_id_rate=10


#do we want to try to mimic interact?
mimic_interact=False


#-----------------------------------------------------------------------------------------------------------------------------
#these parameters can be tuned, but using caution
#-----------------------------------------------------------------------------------------------------------------------------
#agent's ability to change identities - higher means it will shape-shift more
bvagent=0.001
#agent's belief about the client's ability to change identities - higher means it will shape-shift more
bvclient=0.001


#-----------------------------------------------------------------------------------------------------------------------------
#these parameters can be tuned, but will generally work "out of the box" for a basic simulation
#-----------------------------------------------------------------------------------------------------------------------------

#behaviours file
fbfname="fbehaviours.dat"

#identities file
fifname="fidentities.dat"

#get some key parameters from the command line
#set much larger to mimic interact (>5000)
num_samples=500


# use 0.0 for mimic interact simulations
#roughening_noise=0.0
roughening_noise=num_samples**(-1.0/3.0)


#the observation noise
#set to 0.05 to mimic interact
obs_noise=1.0

if mimic_interact:
    bvagent_init=0.000001
    bvclient_init=0.000001
    bvagent=0.00001
    bvclient=0.00001
    agent_knowledge=2
    num_samples=10000
    roughening_noise=0.0
    obs_noise=0.05


#this version does the sampling from a different Gaussian proposal than the actual, then re-weights
learn_meth="rough"

#do we print out all the samples each time
learn_verbose=False

#for repeatability
rseed = NP.random.randint(0,382948932)

print "random seeed is : ",rseed
NP.random.seed(rseed)

#-----------------------------------------------------------------------------------------------------------------------------
#code start - here there be dragons - only hackers should proceed, with caution
#-----------------------------------------------------------------------------------------------------------------------------

#a function to ask the client for an action to take and map it to an EPA value
def ask_client(fbeh,sugg_act='',epaact=[]):
    sst="your action ('?' shows options): "

    sstnc="You can now either :\n"
    sstnc+="- pick an action (behaviour) and type in its label\n"
    sstnc+="- type 'quit' to stop the simulation\n"
    if not sugg_act=='':
        sstnc+="- hit 'enter' to take default action with label: "+sugg_act+"\n"
    if not epaact=='':
        sstnc+="- type any digit (0-9) to take suggsted action: "+str(epaact)

    
    while True:
        cact = raw_input(sst)
        #replace all spaces with _
        cact=re.sub(r"\s+","_",cact.strip())
        if cact=='quit':
            return []
        elif cact=='' and not sugg_act=='':
            cact=sugg_act
            break
        elif cact.isdigit() and not epaact==[]:
            return epaact
        elif cact in fbeh.keys():
            break
        elif cact=="?":
            print sstnc
        else:
            print "not found, try again"
    observ=map(lambda x: float (x), [fbeh[cact]["e"],fbeh[cact]["p"],fbeh[cact]["a"]])
    print "client said: ",cact
    return observ

fbehaviours_agent=readSentiments(fbfname,agent_gender)
fbehaviours_client=readSentiments(fbfname,client_gender)

(agent_mean_ids,agent_cov_ids)=getIdentityStats(fifname,agent_gender)
(client_mean_ids,client_cov_ids)=getIdentityStats(fifname,client_gender)

#the mean and covariance of IDs for male agents as taken from the databases
#should do this automatically in python based on actual genders of client/agent....as above now
#mean_ids=NP.array([0.40760,0.40548,0.45564])
#$cov_ids=NP.array([[2.10735,1.01121, 0.48442],[1.01121,1.22836,0.55593],[0.48442,0.55593,0.77040]])

#the actual (true) ids drawn from the distribution over ids, if not set to something in particular
agent_id=getIdentity(fifname,agent_id,agent_gender)
if agent_id==[]:
    agent_id=NP.random.multivariate_normal(agent_mean_ids,agent_cov_ids)
agent_id=NP.asarray([agent_id]).transpose()
client_id=getIdentity(fifname,client_id,client_gender)
if client_id==[]:
    client_id =  NP.random.multivariate_normal(client_mean_ids,client_cov_ids)
client_id=NP.asarray([client_id]).transpose()

#get initial sets of parameters for agent
(learn_tau_init,learn_prop_init,learn_beta_client_init,learn_beta_agent_init)=init_id(agent_knowledge,agent_id,client_id,client_mean_ids)

#overwrite these values for the interactive script only
#do this for mimicking interact
if mimic_interact:
    learn_beta_agent_init=bvagent_init
    learn_beta_client_init=bvclient_init


#initial x (this is not used for a default agent
learn_initx=NP.array([0.0])


#get the agent - can use some other subclass here if wanted 
learn_agent=Agent(N=num_samples,alpha_value=1.0,initx=learn_initx,
                  gamma_value=obs_noise,beta_value_agent=bvagent,beta_value_client=bvclient,
                  beta_value_client_init=learn_beta_client_init,beta_value_agent_init=learn_beta_agent_init,
                  client_gender=client_gender,agent_gender=agent_gender,
                  agent_rough=roughening_noise,client_rough=roughening_noise)

print 10*"-","learning agent parameters: "
learn_agent.print_params()
print "learner init tau: ",learn_tau_init
print "learner prop init: ",learn_prop_init
print "learner beta client init: ",learn_beta_client_init
print "learner beta agent init: ",learn_beta_agent_init

learn_avgs=learn_agent.initialise_array(learn_tau_init,learn_prop_init,learn_initx)

print "learner average sentiments (f): "
learn_avgs.print_val()

learn_turn="agent"
    

done = False
iter=0
while not done:
    print 10*"#"," current turn: ",learn_turn," ",10*"#"

    observ=[]
    print 10*"-","iter ",iter,80*"-"
    if learn_turn=="agent":
        learn_aab=learn_agent.get_next_action(learn_avgs,learn_turn,meth=learn_meth)
        aact=findNearestBehaviour(learn_aab,fbehaviours_agent)

        print "suggested action for the agent is :",learn_aab,"\n  closest label is: ",aact
        
        learn_aab=ask_client(fbehaviours_client,aact,learn_aab)

        print "agent does action :",learn_aab,"\n"

        learn_observ=[]
    else:
        #first, see what the agent would predict and suggest this to the client
        client_aab=learn_agent.get_next_action(learn_avgs,learn_turn,meth=learn_meth)
        aact=findNearestBehaviours(client_aab,fbehaviours_agent,10)
        print "agent advises the following action :",client_aab,"\n  closest labels are: ", [re.sub(r"_"," ",i.strip()) for i in aact]

        #now, this is where the client actually decides what to do, possibly looking at the suggested labels from the agent
        learn_observ=ask_client(fbehaviours_client,aact[0],client_aab)
        learn_aab=[]
        print "client action: ",learn_observ

    if learn_turn=="client" and learn_observ==[]:
        done = True
    elif learn_turn=="agent" and learn_aab==[]:
        done = True
    else:
        learn_avgs=learn_agent.propagate_forward(learn_turn,learn_aab,learn_observ,meth=learn_meth,verb=learn_verbose)

        print "agent f is: "
        learn_avgs.print_val()

        #I think these should be based on fundamentals, not transients
        (aid,cid)=learn_agent.get_avg_ids(learn_avgs.f)
        print "agent thinks it is most likely a: ",aid
        print "agent thinks the client is most likely a: ",cid

        if get_full_id_rate>0 and (iter+1)%get_full_id_rate==0:
            (cnt_ags,cnt_cls)=learn_agent.get_all_ids()
            print "agent thinks of itself as (full distribution): "
            print cnt_ags[0:10]
            print "agent thinks of the client as (full distribution): "
            print cnt_cls[0:10]
        iter += 1
    print "current deflection of averages: ",learn_agent.deflection_avg
            
    learn_d=learn_agent.compute_deflection()
    print "current deflection (agent's perspective): ",learn_d
    if learn_turn=="client":
        learn_turn="agent"
    elif learn_turn=="agent":
        learn_turn="client"
    

