"""------------------------------------------------------------------------------------------
Bayesian Affect Control Theory
Continuous  Tutor Subclass of Agent class
Author: Jesse Hoey  jhoey@cs.uwaterloo.ca   http://www.cs.uwaterloo.ca/~jhoey
September 2013
Use for research purposes only.
Please do not re-distribute without written permission from the author
Any commerical uses strictly forbidden.
Code is provided without any guarantees.
Research sponsored by the Natural Sciences and Engineering Council of Canada (NSERC).
use python2.6
see README for details
----------------------------------------------------------------------------------------------"""

from bayesact import *


#---------------------------------------------------------------------------------------------------------------------------
#a subclass of Agent that implements some x dynamics
#---------------------------------------------------------------------------------------------------------------------------
class ContinuousTutoringAgent(Agent):
    def __init__(self,*args,**kwargs):

        super(ContinuousTutoringAgent, self).__init__(self,*args,**kwargs)     
        self.x0=kwargs.get("initx",[0.0])
        self.goalx=kwargs.get("goalx",[1.0])
        self.sigma=1.0 #??
        self.delta=0.01 #??


    def print_params(self):
        Agent.print_params(self)
        print "x0: ",self.x0
        print "goal x: ",self.goalx
        print "sigma; ",self.sigma
        print "delta: ",self.delta

    #The only difference between a TutoringAgent and an Agent
    #is dynamics of x and reward
    def initialise_x(self,initx):
        #still default here - to change by adding a bit of noise
        return initx

    def sampleXvar(self,f,tau,x,turn,aab,paab):
        #compute deflection between f and tau
        #x moves towards goalx inversely proportionally to 
        #the deflection
        D=NP.dot(f-tau,f-tau)
        #should be a Gaussian centered around the (deterministic) 
        #effect of aab on x, but with noise dependent on the deflection D
        #NP.random.normal(x[0]+delta_x(x,aab),D)
        #but instead we can do a cheap hack and just move x deterministically
        #towards its goal, and then add noise that scales it back proportional to D
        #new_x=x[0]+self.delta*(1.0-min(1.0,NP.random.poisson((D+1.0)/self.sigma)/20.0))
	new_x=NP.random.normal(x[0]+0.01/D,1.0)
        return [new_x]

    def reward(self,sample):
        # a generic deflection-based reward that can be removed eventually
        fsample=sample.f.transpose()
        reward = -1.0*NP.dot(sample.f-sample.tau,sample.f-sample.tau)
        # a state-based reward
        reward = -(sample.x - self.goalx)**2
        return reward



